;************************************************************************************
; Commodore 64 emulator
;
; Translated and corrected sourcecodes
;
; originally  ED64 - HOW TO WRITE A COMMODORE 64 EMULATOR
;		  Copyright 2007 ir. Marc Dendooven
;
; Blitz3D Conversion by Jimmy
;
; bugfixed to a running state and commented by Lee Davison 2011/06/05
;
; visit:  http://www.themotionstore.com/leeedavison/
;
; Sorry, the text is formatted for TAB stops at six characters not four characters
; as used by the Blitz IDE but the Blitz syntax highlighting colours make me want to
; stab my eyes out. So I used another, monochrome, text editor.
;
; Version 0.63
;
;	Fixed the keyboard shift sense but shift lock still does nothing.
;
; Version 0.62
;
;	Replaced the pixel by pixel drawing of characters on the C64 screen with tiles
;	loaded from an image. This makes drawing a whole screen of characters, as is
;	done whenever the screen scrolls, five or six times faster than previously.
;
; Version 0.60
;
;	A fairly major speedup has been achieved by replacing the arrays for ram and
;	the three roms with banks. this makes loading the roms much simpler and reading
;	memory or writing ram quicker. It now runs at under 100% CPU loading on my old
;	laptop even in debug mode. Writing to the screen still causes a huge slowdown
;	though. There must be some way to use the bitmap from the character memory
;	directly.
;
;	the jmp() function has been replaced by commands in the case statement
;
; Version 0.59
;
;	Added the showtime constant and some extra code at the end of the main loop
;	that will show how many milliseconds late the code was in completing 20000 CPU
;	cycles. To see this set the showtime constant to True.
;
; Version 0.58
;
;	Corrected the cycle timings for the IRQ and NMI interrupts.
;
;	Corrected the cycle timings for most opcodes and address modes except for
;	some cases of address,{X|Y} indexing. The error is only one cycle at most.
;
;	This means that on a fast enough machine this now runs at about the same
;	speed as a real Commodore 64.
;
;	reset() now clears the cycle counter.
;
; Version 0.57
;
;	Replaced the busy wait at the end of the main loop with a delay wait. This
;	means a lower CPU load on fast machines and faster execution on slow machines.
;
; Version 0.56
;
;	Removed a spurious Int() from the VicText()function.
;
; Version 0.55
;
;	Removed the Goto from the irq function.
;
; Version 0.54
;
;	Added the correct attribution.
;
;************************************************************************************
;
; Emulates a 6502 (without secret opcodes and extra 6510 opcodes, just standard C64
; opcodes)
; Does not emulate SID or CIA(input/output 2 6526 adapters w16bit parallel IO 8bit
; seraial IO, TOD clock with alarm)
; Does an crudly simple VIC II emulation (40x25 textmode 1024 area without colourmap,
; and vbl interrupt) nothing else
; 2Joystickports, userport, cartridge, disk, datasette

; Improved but as of yet unused C64 colour palette:
;
; $000000	black
; $FFFFFF	white
; $68372B	red
; $70A4B2	cyan
; $6F3D86	purple
; $588D43	green
; $352879	blue
; $B8C76F	yellow
; $6F4F25	lt brown
; $433900	brown
; $9A6759	pink
; $444444	dark grey
; $6C6C6C	grey
; $9AD284	lt green
; $6C5EB5	lt blue
; $959595	lt grey

;************************************************************************************
;
; Constants

Const n = $80					; status bits in p: NV..DIZC
Const v = $40
Const b = $10					; b isn't a real bit, don't set it
Const d = $08
Const i = $04
Const z = $02
Const c = $01

Const showtime = False				; if you want to see how much time is being
							; taken to do 20000 CPU cycles then set
							; this to True

;************************************************************************************
;
; Global variables

							; 8 bit registers
Global a						; Accumulator
Global x						; X register
Global y						; Y register
Global s						; the low 8 bits of the stack pointer
Global p						; processor status register
Global op						; the current opcode

							; 16 bit register
Global pc						; program counter

Global cycle					; processor cycle counter

Global ram						; 64K RAM bank
Global kernal					;  8K RAM bank
Global basic					;  8K RAM bank
Global char						;  4K RAM bank
Global tiles					; character tiles

ram = CreateBank(65536)				; $0000-$FFFF
kernal = CreateBank($2000)			; $E000-$FFFF
basic = CreateBank($2000)			; $A000-$BFFF
char = CreateBank($1000)			; $D000-$DFFF


;************************************************************************************
;
; start of code

Graphics 640,400					; a big enough window for no borders

; Load the roms

file = ReadFile("Kernal.rom")			; load the kernal ROM, 8K
ReadBytes kernal,file,0,$2000
CloseFile file

file = ReadFile("basic.rom")			; load the basic ROM, 8K
ReadBytes basic,file,0,$2000
CloseFile file


file = ReadFile("char.rom")			; load the character ROM, 4K
ReadBytes char,file,0,$1000
CloseFile file

tiles = LoadAnimImage("chars.png",16,16,0,256)
							; load the charater tiles, first 250 only

; start up the processor

reset()						; set the initial pc and disable interrupts
end_time = MilliSecs() + 20			; set the initial interrupt time

; Main loop

Repeat
  cycle = cycle + 2				; all instructions are a minimum of two cycles
  op = peek(pc)					; get the next opcode
  pc = (pc + 1) And $FFFF			; increment the program counter

  Select op
	Case $69 : adc(imm())			; was broken, now ok
	Case $65 : adc(zp())			; was broken, now ok
	Case $75 : adc(zpx())			; was broken, now ok
	Case $6D : adc(abs_())			; was broken, now ok
	Case $7D : adc(abs_x())			; was broken, now ok
	Case $79 : adc(abs_y())			; was broken, now ok
	Case $61 : adc(indx())			; was broken, now ok
	Case $71 : adc(indy())			; was broken, now ok
	Case $29 : and_(imm())			; ok
	Case $25 : and_(zp())			; ok
	Case $35 : and_(zpx())			; ok
	Case $2D : and_(abs_())			; ok
	Case $3D : and_(abs_x())		; ok
	Case $39 : and_(abs_y())		; ok
	Case $21 : and_(indx())			; ok
	Case $31 : and_(indy())			; ok
	Case $0A : asl_A()			; ok
	Case $06 : asl(zp())			; ok
	Case $16 : asl(zpx())			; ok
	Case $0E : asl(abs_())			; ok
	Case $1E : asl(abs_x())			; ok
	Case $90 : bfc(c)				; ok ; bcc
	Case $B0 : bfs(c)				; ok ; bcs
	Case $F0 : bfs(z)				; ok ; beq
	Case $24 : bit(zp())			; was broken, now ok
	Case $2C : bit(abs_())			; was broken, now ok
	Case $30 : bfs(n)				; ok ; bmi
	Case $D0 : bfc(z)				; ok ; bne
	Case $10 : bfc(n)				; ok ; bpl
	Case $00 : brk()				; ok
	Case $50 : bfc(v)				; ok ; bvc
	Case $70 : bfs(v)				; ok ; bvs
	Case $18 : setflag(c,False)		; ok ; clc
	Case $D8 : setflag(d,False)		; ok ; cld
	Case $58 : setflag(i,False)		; ok ; cli
	Case $B8 : setflag(v,False)		; ok ; clv
	Case $C9 : cmp(imm())			; ok
	Case $C5 : cmp(zp())			; ok
	Case $D5 : cmp(zpx())			; ok
	Case $CD : cmp(abs_())			; ok
	Case $DD : cmp(abs_x())			; ok
	Case $D9 : cmp(abs_y())			; ok
	Case $C1 : cmp(indx())			; ok
	Case $D1 : cmp(indy())			; ok
	Case $E0 : cpx(imm())			; ok
	Case $E4 : cpx(zp())			; ok
	Case $EC : cpx(abs_())			; ok
	Case $C0 : cpy(imm())			; ok
	Case $C4 : cpy(zp())			; ok
	Case $CC : cpy(abs_())			; ok
	Case $C6 : dec(zp())			; ok
	Case $D6 : dec(zpx())			; ok
	Case $CE : dec(abs_())			; ok
	Case $DE : dec(abs_x())			; ok
	Case $CA : dex()				; ok
	Case $88 : dey()				; ok
	Case $49 : eor(imm())			; ok
	Case $45 : eor(zp())			; ok
	Case $55 : eor(zpx())			; ok
	Case $4D : eor(abs_())			; ok
	Case $5D : eor(abs_x())			; ok
	Case $59 : eor(abs_y())			; ok
	Case $41 : eor(indx())			; ok
	Case $51 : eor(indy())			; ok
	Case $E6 : inc(zp())			; ok
	Case $F6 : inc(zpx())			; ok
	Case $EE : inc(abs_())			; ok
	Case $FE : inc(abs_x())			; ok
	Case $E8 : inx()				; ok
	Case $C8 : iny()				; ok
	Case $4C : cycle = cycle - 1		; correction for no additional memory read
		     pc = abs_()			; ok ; jmp
	Case $6C : pc = ind()			; ok ; jmp
	Case $20 : jsr(abs_())			; ok
	Case $A9 : lda(imm())			; ok
	Case $A5 : lda(zp())			; ok
	Case $B5 : lda(zpx())			; ok
	Case $AD : lda(abs_())			; ok
	Case $BD : lda(abs_x())			; ok
	Case $B9 : lda(abs_y())			; ok
	Case $A1 : lda(indx())			; ok
	Case $B1 : lda(indy())			; ok
	Case $A2 : ldx(imm())			; ok
	Case $A6 : ldx(zp())			; ok
	Case $B6 : ldx(zpy())			; ok
	Case $AE : ldx(abs_())			; ok
	Case $BE : ldx(abs_y())			; ok
	Case $A0 : ldy(imm())			; ok
	Case $A4 : ldy(zp())			; ok
	Case $B4 : ldy(zpx())			; ok
	Case $AC : ldy(abs_())			; ok
	Case $BC : ldy(abs_x())			; ok
	Case $4A : lsr_A()			; ok
	Case $46 : lsr(zp())			; ok
	Case $56 : lsr(zpx())			; ok
	Case $4E : lsr(abs_())			; ok
	Case $5E : lsr(abs_x())			; ok
	Case $EA : ; nop				; ok
	Case $09 : ora(imm())			; ok
	Case $05 : ora(zp())			; ok
	Case $15 : ora(zpx())			; ok
	Case $0D : ora(abs_())			; ok
	Case $1D : ora(abs_x())			; ok
	Case $19 : ora(abs_y())			; ok
	Case $01 : ora(indx())			; ok
	Case $11 : ora(indy())			; ok
	Case $48 : pha()				; ok
	Case $08 : php()				; was broken, now ok
	Case $68 : pla()				; ok
	Case $28 : plp()				; was broken, now ok
	Case $2A : rol_A()			; ok
	Case $26 : rol(zp())			; ok
	Case $36 : rol(zpx())			; ok
	Case $2E : rol(abs_())			; ok
	Case $3E : rol(abs_x())			; ok
	Case $6A : ror_A()			; ok
	Case $66 : ror(zp())			; ok
	Case $76 : ror(zpx())			; ok
	Case $6E : ror(abs_())			; ok
	Case $7E : ror(abs_x())			; ok
	Case $40 : rti()				; ok
	Case $60 : rts()				; ok
	Case $E9 : sbc(imm())			; was broken, now ok
	Case $E5 : sbc(zp())			; was broken, now ok
	Case $F5 : sbc(zpx())			; was broken, now ok
	Case $ED : sbc(abs_())			; was broken, now ok
	Case $FD : sbc(abs_x())			; was broken, now ok
	Case $F9 : sbc(abs_y())			; was broken, now ok
	Case $E1 : sbc(indx())			; was broken, now ok
	Case $F1 : sbc(indy())			; was broken, now ok
	Case $38 : setflag(c,True)		; ok ; sec
	Case $F8 : setflag(d,True)		; ok ; sed
	Case $78 : setflag(i,True)		; ok ; sei
	Case $85 : sta(zp())			; ok
	Case $95 : sta(zpx())			; ok
	Case $8D : sta(abs_())			; ok
	Case $9D : sta(abs_x())			; ok
	Case $99 : sta(abs_y())			; ok
	Case $81 : sta(indx())			; ok
	Case $91 : sta(indy())			; ok
	Case $86 : stx(zp())			; ok
	Case $96 : stx(zpy())			; ok
	Case $8E : stx(abs_())			; ok
	Case $84 : sty(zp())			; ok
	Case $94 : sty(zpx())			; ok
	Case $8C : sty(abs_())			; ok
	Case $AA : tax()				; ok
	Case $A8 : tay()				; ok
	Case $BA : tsx()				; ok
	Case $8A : txa()				; ok
	Case $9A : txs()				; ok
	Case $98 : tya()				; ok
	Case $3F					; use opcode $3F to drop into the debugger
	  Stop
	Case $7F					; use opcode $7F to output a value
	  Print peek(pc)
	  pc = (pc + 1) And $FFFF
	Default					; an invalid opcode tells you where it is
	  Write pc - 1 + " " + op
	  Print " unknown instruction"
	  WaitKey
	End
  End Select

; 1000000 cycles/second * 20ms vertical blanking. doing the timing this way means that
; no time is wasted waiting for the next vertical blank if more than 20ms have passed
; since the last 20000 cycles. it also means that on fast enough machines the CPU
; loading is significantly reduced. on my machine with debug turned off the CPU load
; goes from 100% to 65%

  If cycle > 20000 Then				; if at least 20000 cycles have been done
    cycle = cycle - 20000			; decrement the cycle count
    wait_time = end_time - MilliSecs()	; get the time to wait for the next 20ms
							; interval
    If wait_time > 0 Then			; if there is time to wait
	Delay wait_time				;   wait that time
    End If

; if you want to know how long it is taking to execute those 20000 CPU cycles then
; set showtime to True and a small box at the top left of the screen will tell you
; by how many milliseconds it missed finishing in under 20 milliseconds.
;
; So whatever number it shows it took 20ms plus that number to do those 20000 cycles.

    If showtime Then				; if the user wants to see the time
      Color 0,0,0					; black
      Rect 0,1, 23,12, True			; clear the rectangle
      Color $FF,$FF,$FF				; white
      Locate 0,0					; top left of screen
      Write -wait_time				; write the time
    End If

    irq()						; 20ms is up so fire the interrupt request
    end_time = MilliSecs() + 20		; set the next interrupt time
  End If

Until MouseDown(2)
End

;************************************************************************************
;
; functions that read/write memory

; return a word from memory

Function peek2(address)
  If address=>$E000 Then			; kernal ROM address space
    peek2 = PeekShort(kernal,address-$E000)
  Else If address=>$D000 Then			; I/O address space
    peek2 = peek(address) + peek(address + 1) Shl 8
  Else If address=>$C000 Then			; RAM address space
    peek2 = PeekShort(ram,address)
  Else If address=>$A000 Then			; basic ROM address space
    peek2 = PeekShort(basic,address-$A000)
  Else						; RAM address space
    peek2 = PeekShort(ram,address)
  End If
  Return peek2
End Function


; return a byte from memory

Function peek(address)
  If address=>$E000 Then			; kernal ROM address space
    peek = PeekByte(kernal,address-$E000)
  Else If address=>$D000 Then			; I/O address space
    peek = io(address)
  Else If address=>$C000 Then			; RAM address space
    peek = PeekByte(ram,address)
  Else If address=>$A000 Then			; basic ROM address space
    peek = PeekByte(basic,address-$A000)
  Else						; RAM address space
    peek = PeekByte(ram,address)
  End If
  Return peek
End Function


; return a byte from an i/o device

Function io(address)
  If address = $D012 Then			 ; fake a raster line
    io = $00
  Else
    io = $FF
  End If
  Return io
End Function


; poke a byte into memory

Function poke(address,value)
  PokeByte ram,address,value
  If (address=>$0400) And (address=<$07E7) Then VicText(address-$0400,value)
End Function


;************************************************************************************
;
; functions which return address for the instructions to work with

; return the address of the immediate byte

Function imm()	
  imm = pc						; the byte follows the opcode
  pc = (pc + 1) And $FFFF
  Return imm
End Function


; return the address of the zero page byte

Function zp()
  cycle = cycle + 1				; address mode timing addition
  zp = peek(pc)
  pc = (pc + 1) And $FFFF
  Return zp
End Function


; return the address of the zero page byte indexed by X

Function zpx()
  cycle = cycle + 2				; address mode timing addition
  zpx = (peek(pc) + x) And $FF
  pc = (pc + 1) And $FFFF
  Return zpx
End Function


; return the address of the zero page byte indexed by Y

Function zpy()
  cycle = cycle + 2				; address mode timing addition
  zpy = (peek(pc) + y) And $FF
  pc = (pc + 1) And $FFFF
  Return zpy
End Function


; return the absolute address

Function abs_()
  cycle = cycle + 2				; address mode timing addition
  abs_ = peek2(pc)
  pc = (pc + 2) And $FFFF
  Return abs_
End Function


; return the absolute address indexed by X

Function abs_x()
  cycle = cycle + 2				; address mode timing addition
  absx = (peek2(pc) + x) And $FFFF
  pc = (pc + 2) And $FFFF
  Return absx
End Function


; return the absolute address indexed by Y

Function abs_y()
  cycle = cycle + 2				; address mode timing addition
  absy = (peek2(pc) + y) And $FFFF
  pc = (pc + 2) And $FFFF
  Return absy
End Function


; return the indirect address, this is broken on the NMOS 6502 if it crosses a page
; boundary

Function ind()
  cycle = cycle + 3				; address mode timing addition
  ind = peek2(peek2(pc))
  pc = (pc + 2) And $FFFF
  Return ind
End Function


; return the (indirect,X) address

Function indx()
  cycle = cycle + 4				; address mode timing addition
  indx = peek2((peek(pc) + x) And $FF)
  pc = (pc + 1) And $FFFF
  Return indx
End Function


; return the (indirect),Y address

Function indy()
  cycle = cycle + 3				; address mode timing addition
  indy = (peek2(peek(pc)) + y) And $FFFF
  pc = (pc + 1) And $FFFF
  Return indy
End Function


;************************************************************************************
;
; flags, bits, words and bytes

; set the flag bit in the status byte
;
; never use this routine to set the b flag as it won't get cleared and everything will
; then go pear shaped

Function setflag(flag,status)
  If status Then					; if status is non zero
    p = (p Or flag)				;   set the flag bit
  Else						; else
    p = (p And (flag Xor $FF))		;   clear the flag bit
  End If
End Function


; return the 0 or 1 state of a flag bit

Function flagset(flag)
  flagset = (p And flag) <> 0			; the <> 0 ensures only 0 or 1 is returned
  Return flagset
End Function


; return the low byte of a word

Function lo(value)
  Return value And $FF
End Function


; return the high byte of a word

Function hi(value)
  value = (value Shr 8) And $FF
  Return value
End Function


; push a byte onto the stack and decrement the stack pointer

Function push(byte)
  poke($100 + s,byte)
  s = (s - 1) And $FF
End Function


; increment the stack pointer and pull a byte from the stack

Function pull()
  s = (s + 1) And $FF
  pull = peek($100 + s)
  Return pull
End Function


; sign extend a byte to 32 bits

Function signedbyte(byte)
  h = byte
  If h => 128 Then h = h - $100
  Return h
End Function


;************************************************************************************
;
; the opcodes

; branch if the flag bit is set

Function bfs(flag)
  If flagset(flag) Then branch()		; if the flag is set take the branch
  pc = (pc + 1) And $FFFF
End Function


; branch if the flag bit is clear

Function bfc(flag)
  If Not flagset(flag) Then branch()	; if the flag is not set take the branch
  pc = (pc + 1) And $FFFF
End Function


; take the branch

Function branch()
  cycle = cycle + 1				; branch taken timing addition
  page = (pc + 1) And $FF00			; copy this page
  pc = (pc + signedbyte(peek(pc))) And $FFFF
							; calculate and mask the destination address
  If page <> ((pc + 1) And $FF00) Then	; check for page crossing
    cycle = cycle + 1				;   page crossed timing addition
  End If
End Function


; decimal mode is not needed to run the C64 firmware

Function adc(address)
  val = peek(address)				; get the value from memory
  bit = flagset(c)				; get the carry value in bit
  hc = (((a And $7F) + (val And $7F) + bit) And $80) = $80
							; calculate the carry from b6 to b7
  h = a + val + bit				; calculate the new value and new carry
  a = h And $FF					; and copy the value part to A
  h = (h And $100) = $100			; calculate the carry from b7
  setflag(c,h)					; set the carry bit in the status
  setflag(z,Not a)				; set the zero bit in the status
  setflag(n,a And n)				; set the negative bit in the status
  setflag(v,h Xor hc)				; set the overflow bit in the status
							; overflow is true when hc and h differ
End Function


; decimal mode is not needed to run the C64 firmware

Function sbc(address)
  val = peek(address)				; get the value from memory
  bit = Not flagset(c)				; get the inverted carry value in bit
  hc = Not (((a And $7F) - (val And $7F) - bit) And $80)
							; calculate the inverted carry from b6 to b7
  h = a - val - bit				; calculate the new value and new carry
  a = h And $FF					; and copy the value part to A
  h = Not (h And $100)				; calculate the inverted carry from b7
  setflag(c,h)					; set the carry bit in the status
  setflag(z,Not a)				; set the zero bit in the status
  setflag(n,a And n)				; set the negative bit in the status
  setflag(v,h Xor hc)				; set the overflow bit in the status
							; overflow is true when hc and h differ
End Function


Function asl(address)
  cycle = cycle + 2				; read/modify/write timing addition
  byte = peek(address) Shl 1			; get and shift the value
  setflag(c,byte And $100)			; set the carry bit in the status
  byte = byte And $FF				; mask the result to byte size
  poke(address,byte)				; save the result back to memory
  setflag(z,Not byte)				; set the zero bit in the status
  setflag(n,byte And n)				; set the negative bit in the status
End Function


Function asl_A()
  a = a Shl 1					; shift the value
  setflag(c,a And $100)				; set the carry bit in the status
  a = a And $FF					; mask the result to byte size
  setflag(z,Not a)				; set the zero bit in the status
  setflag(n,a And n)				; set the negative bit in the status
End Function


Function rol(address)
  cycle = cycle + 2				; read/modify/write timing addition
  byte = (peek(address) Shl 1) + flagset(c)
							; get shift and add the carry to the value
  setflag(c,byte And $100)			; set the carry bit in the status
  byte = byte And $FF				; mask the result to byte size
  poke(address,byte)				; save the result back to memory
  setflag(z,Not byte)				; set the zero bit in the status
  setflag(n,byte And n)				; set the negative bit in the status
End Function


Function rol_A()
  a = (a Shl 1) +  flagset(c)			; shift and add the carry to the value
  setflag(c,a And $100)				; set the carry bit in the status
  a = a And $FF					; mask the result to byte size
  setflag(z,Not a)				; set the zero bit in the status
  setflag(n,a And n)				; set the negative bit in the status
End Function


Function ror(address)
  cycle = cycle + 2				; read/modify/write timing addition
  byte = peek(address)				; get the value from memory
  bit = flagset(c) Shl 7			; get the carry value in b7
  setflag(c,byte And 1)				; set the carry bit in the status
  byte = (byte Shr 1) Or bit			; shift and add the carry to the value
  poke(address,byte)				; save the result back to memory
  setflag(z,Not byte)				; set the zero bit in the status
  setflag(n,byte And n)				; set the negative bit in the status
End Function


Function ror_A()
  bit = flagset(c) Shl 7			; get the carry value in b7
  setflag(c,a And 1)				; set the carry bit in the status
  a = (a Shr 1) Or bit				; shift and add the carry to the value
  setflag(z,Not a)				; set the zero bit in the status
  setflag(n,a And n)				; set the negative bit in the status
End Function


Function lsr(address)
  cycle = cycle + 2				; read/modify/write timing addition
  byte = peek(address)				; get the value from memory
  setflag(c,byte And 1)				; set the carry bit in the status
  byte = byte Shr 1				; shift the value
  poke(address,byte)				; save the result back to memory
  setflag(z,Not byte)				; set the zero bit in the status
  setflag(n,False)				; clear the negative bit in the status
End Function


Function lsr_A()
  setflag(c,a And 1)				; set the carry bit in the status
  a = a Shr 1					; shift the value
  setflag(z,Not a)				; set the zero bit in the status
  setflag(n,False)				; clear the negative bit in the status
End Function


Function bit(address)
  h = peek(address)				; get the value from memory
  setflag(n,h And n)				; set the negative bit in the status
  setflag(v,h And v)				; set the overflow bit in the status
  setflag(z,Not(h And a))			; set the zero bit in the status
End Function


Function brk()
  cycle = cycle + 5				; break timing addition
  pc = (pc + 1) And $FFFF			; increment the pc
  push(hi(pc))					; push the pc high byte
  push(lo(pc))					; push the pc low byte
  push(p Or b)					; push the status byte with the break bit set
  setflag(i,True)					; set the interrupt disable bit
  pc = peek2($FFFE)				; get the IRQ vector
End Function


Function and_(address)
  a = a And peek(address)			; AND the address contents with A
  setflag(z,Not a)				; set the zero bit in the status
  setflag(n,a And n)				; set the negative bit in the status
End Function


Function ora(address)
  a=a Or peek(address)				; OR the address contents with A
  setflag(z,Not a)				; set the zero bit in the status
  setflag(n,a And n)				; set the negative bit in the status
End Function


Function eor(address)
  a = a Xor peek(address)			; EOR the address contents with A
  setflag(z,Not a)				; set the zero bit in the status
  setflag(n,a And n)				; set the negative bit in the status
End Function


Function cmp(address)
  h = a - peek(address)				; calculate A = the address contents
  setflag(c,Not (h And $100))			; set the carry bit in the status
  setflag(z,Not (h And $FF))			; set the zero bit in the status
  setflag(n,h And n)				; set the negative bit in the status
End Function


Function cpx(address)
  h = x - peek(address)				; calculate X = the address contents
  setflag(c,Not (h And $100))			; set the carry bit in the status
  setflag(z,Not (h And $FF))			; set the zero bit in the status
  setflag(n,h And n)				; set the negative bit in the status
End Function


Function cpy(address)
  h = y - peek(address)				; calculate Y = the address contents
  setflag(c,Not (h And $100))			; set the carry bit in the status
  setflag(z,Not (h And $FF))			; set the zero bit in the status
  setflag(n,h And n)				; set the negative bit in the status
End Function


Function dec(address)
  cycle = cycle + 2				; read/modify/write timing addition
  h = (peek(address) - 1) And $FF		; calculate the address contents - 1
  poke(address,h)					; save the result back to memory
  setflag(z,Not h)				; set the zero bit in the status
  setflag(n,h And n)				; set the negative bit in the status
End Function


Function inc(address)
  cycle = cycle + 2				; read/modify/write timing addition
  h = (peek(address) + 1) And $FF		; calculate the address contents + 1
  poke(address,h)					; save the result back to memory
  setflag(z,Not h)				; set the zero bit in the status
  setflag(n,h And n)				; set the negative bit in the status
End Function


Function dex()
  x = (x - 1) And $FF				; calculate X - 1
  setflag(z,Not x)				; set the zero bit in the status
  setflag(n,x And n)				; set the negative bit in the status
End Function


Function inx()
  x = (x + 1) And $FF				; calculate X + 1
  setflag(z,Not x)				; set the zero bit in the status
  setflag(n,x And n)				; set the negative bit in the status
End Function


Function dey()
  y = (y - 1) And $FF				; calculate Y - 1
  setflag(z,Not y)				; set the zero bit in the status
  setflag(n,y And n)				; set the negative bit in the status
End Function


Function iny()
  y = (y + 1) And $FF				; calculate Y + 1
  setflag(z,Not y)				; set the zero bit in the status
  setflag(n,y And n)				; set the negative bit in the status
End Function


Function jsr(address)
  cycle = cycle + 2				; address push timing addition
  pc = (pc - 1) And $FFFF			; decrement the address
  push(hi(pc))					; push the pc high byte
  push(lo(pc))					; push the pc low byte
  pc = address					; copy the address to the pc
End Function


Function rts()
  cycle = cycle + 4				; address pull timing addition
  pc = 1 + pull()					; pull the pc low byte
  pc = pc + pull() Shl 8			; pull the pc high byte
End Function


Function rti()
  cycle = cycle + 4				; address pull timing addition
  p = pull()					; pull the status byte
  pc = pull()					; pull the pc low byte
  pc = pc + pull() Shl 8			; pull the pc high byte
End Function


Function lda(address)
  a = peek(address)				; get the value from memory
  setflag(z,Not a)				; set the zero bit in the status
  setflag(n,a And n)				; set the negative bit in the status
End Function


Function ldx(address)
  x = peek(address)				; get the value from memory
  setflag(z,Not x)				; set the zero bit in the status
  setflag(n,x And n)				; set the negative bit in the status
End Function


Function ldy(address)
  y = peek(address)				; get the value from memory
  setflag(z,Not y)				; set the zero bit in the status
  setflag(n,y And n)				; set the negative bit in the status
End Function


Function pha()
  cycle = cycle + 1				; byte push timing addition
  push(a)						; push a
End Function


Function pla()
  cycle = cycle + 2				; byte pull timing addition
  a = pull()					; pull a byte from the stack
  setflag(z,Not a)				; set the zero bit in the status
  setflag(n,a And n)				; set the negative bit in the status
End Function


Function php()
  cycle = cycle + 1				; byte push timing addition
  push(p Or b)					; push the status byte with the break bit set
End Function


Function plp()
  cycle = cycle + 2				; byte pull timing addition
  p = pull() AND (b Xor $FF)			; pull the status byte and clear the break bit
End Function


Function sta(address)
  poke(address,a)					; save A to the memory
End Function


Function stx(address)
  poke(address,x)					; save X to the memory
End Function


Function sty(address)
  poke(address,y)					; save Y to the memory
End Function


Function tax()
  x = a						; copy X to A
  setflag(z,Not x)				; set the zero bit in the status
  setflag(n,x And n)				; set the negative bit in the status
End Function


Function tay()
  y = a						; copy Y to A
  setflag(z,Not y)				; set the zero bit in the status
  setflag(n,y And n)				; set the negative bit in the status
End Function


Function tsx()
  x = s						; copy X to the stack pointer
  setflag(z,Not x)				; set the zero bit in the status
  setflag(n,x And n)				; set the negative bit in the status
End Function


Function txa()
  a = x						; copy A to X
  setflag(z,Not a)				; set the zero bit in the status
  setflag(n,a And n)				; set the negative bit in the status
End Function


Function txs()
  s = x						; copy the stack pointer to X
End Function


Function tya()
  a = y						; copy A to Y
  setflag(z,Not a)				; set the zero bit in the status
  setflag(n,a And n)				; set the negative bit in the status
End Function


;************************************************************************************
;
; the minimal hardware emulation

; interrupts

Function nmi()
  interrupt($FFFA)				; do the NMI interrupt
End Function


Function reset()
  cycle = 0						; clear the cycle counter
  setflag(i,True)					; set the interrupt disable bit
  pc = peek2($FFFC)				; get the RESET vector
End Function


; vertical blanking interrupt

Function irq()
  If Not flagset(i) Then			; if the disable bit isn't set
    interrupt($FFFE)				;   do the IRQ interrupt
    addkey()					;   pretend to scan the keyboard
  End If
End Function


; generic interrupt

Function interrupt(vector)
  cycle = cycle + 7				;   interrupt timing addition
  push(hi(pc))					;   push the pc high byte
  push(lo(pc))					;   push the pc low byte
  push(p)						;   push the status byte with break bit clear
  setflag(i,True)					;   set the interrupt disable bit
  pc = peek2(vector)				;   get the interrupt vector
End Function


; draw a character tile on the screen this much faster than pixel by pixel

Function vicText(pos,val)
  x0 = (pos Mod 40) * 16			; left edge x co-ordinate from pos
  y0 = (pos / 40) * 16				; top edge y co-ordinate from pos
  DrawImage tiles,x0,y0,val			; draw the character
End Function


; keyboard routine. shift lock doesn't seem to work and shift is the wrong sense

Function addkey()
  kb = GetKey() And $FF				; get a key
  If kb Then					; if there is a key
    If ((kb>$40) And (kb<$5B)) Or ((kb>$60) And (kb<$7B)) Then
      kb = kb Xor $20				; swap upper and lower case
    End If
    buff = peek($C6)				;   get the keyboard buffer index
    If buff < peek($0289) Then		;   if there's room in the buffer
      poke($0277 + buff,kb)			;     save the key
      poke($C6,buff + 1)			;     save the updated buffer index
    End If
  End If
End Function

;************************************************************************************
